/***
 * Excerpted from "No Fluff Just Stuff 2006 Anthology"
 * We make no guarantees that this code is fit for any purpose. 
 * Visit http://www.pragmaticprogrammer.com/titles/nfjs06 for more book information.
***/
package net.bosatsu.util.regex.io;

import java.io.BufferedReader;
import java.io.FilterInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;

/**
 *
 * This class demonstrates the Decorator Pattern use in the JDK I/O classes
 * by allowing a developer to add regular expression substitution to an underlying
 * InputStream. This implementation deals with single-line patterns. Multiple patterns
 * may be applied to each line of the input, and each successful match on a line will
 * result in the replacement of the matched text with the corresponding substitution.
 * We assume that the InputStream being wrapped is a text-oriented stream capable of
 * being read a line at a time.
 *
 * @author bsletten
 * @author dsletten
 *
 * @see net.bosatsu.util.regex.io.RegularExpressionInputStream
 * @see net.bosatsu.util.regex.io.RegularExpressionReplacer
 *
 */

public class RegularExpressionInputStream extends FilterInputStream {
    public static final int EOF = -1;
    public static final int EOL = '\n';

    private BufferedReader br;
    private RegularExpressionReplacer rer;

    private byte[] buffer;
    private int bufferIndex;

    /*
     * Main constructor takes the InputStream to decorate and a RegularExpressionReplacer
     * instance which captures the correspondence between regular expression patterns
     * and the Transformers to apply when those patterns succeed.
     *
     */
    public RegularExpressionInputStream( InputStream is, RegularExpressionReplacer rer ) throws IOException
    {
        super( is );
        br = new BufferedReader( new InputStreamReader( in ) );
        this.rer = rer;

        // Populate the buffer with a line, transforming
        // if there are any matches
        fetchAndTransformNextLine();
    }

    /*
     * Read the next byte from the stream.
     *
     */
    public int read() throws IOException {
        int retValue = EOF;

        // If we have a buffer
        if( buffer != null ) {
            // And it isn't empty
            if( bufferIndex < buffer.length ) {
                // Read the next byte
                retValue = buffer[ bufferIndex++ ];
            } else {
                // We've hit the end of our line.
                // Grab the next line and transform
                // anything that matches any of our
                // patterns.
                fetchAndTransformNextLine();

                // We need to indicate the EOL
                retValue = EOL;
            }
        }

        return retValue;
    }

    /*
     * Try to read into the buffer.
     *
     */
    public int read( byte[] b ) throws IOException {
        return read( b, 0, b.length );
    }

    /*
     *
     * Try to read into the buffer.
     *
     */

    public int read( byte[] b, int offset, int len ) throws IOException {

        int retValue = EOF;
        int numRead = 0;

        boolean done = false;
        int idx = offset;

        while( !done && ( numRead < len ) ) {
            int nextByte = read();

            if( nextByte != EOF ) {
                b[ idx++ ] = ( byte ) nextByte;
                numRead++;
            } else {
                done = true;
                retValue = EOF;
            }
        }

        // If we didn't fall off the end indicate how many
        // bytes we got from the stream

        if( numRead > 0 ) {
            retValue = numRead;
        }

        return retValue;
    }

    /*
     * Read in the next line from the BufferedReader and transform
     * it as necessary before making it available for reading.
     *
     */
    private void fetchAndTransformNextLine() throws IOException {
        String line = br.readLine();

        bufferIndex = 0;

        if( line != null ) {
            buffer = rer.performTransformations( line ).getBytes();
        } else {
            buffer = null;
        }
    }

    /*
     * Close the BufferedReader.
     */

    public void close() throws IOException {
        br.close();
    }

    /*
     * Return the number of bytes available in the transformed
     * buffer.
     */

    public int available() {
        return ( buffer == null )
            ? 0 : buffer.length - bufferIndex;
    }

    /*
     * Skip over the specified bytes
     */

    public long skip( long n ) throws IOException {
        long retValue = 0;

        int availableBytes = available();

        if( availableBytes >= n ) {
            bufferIndex += n;
            retValue = n;
        } else {
            long num = n;

            while( ( read() != EOF ) && num > 0 );

            retValue = n - num;
        }

        return retValue;
    }

    /*
     * We don't support mark() and reset() because
     * we are transforming the stream.
     */

    public void reset() throws IOException {
        throw new UnsupportedOperationException();
    }

    public boolean markSupported() {
        return false;
    }
}

